/************************************************************************

Licensed Materials - Property of IBM

IBM Surveillance Insights for Financial Services

© Copyright IBM Corp. 2017

US Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

************************************************************************/


var NodeProp = {}; //store properties values for the graph
NodeProp.svg = null; //dom svg object
NodeProp.simulation = null; //d3 forceSimulation object
NodeProp.evidenceScore = null; //evidence score for specific date
NodeProp.graphData = null; //graph data
NodeProp.width = null;
NodeProp.height = 500;
NodeProp.isDrillDown = false; //boolean to store drill down status
NodeProp.drillDownNode = false; //store select node for drilldown
NodeProp.isInitialStateGenerated = false;
NodeProp.initialState = {
    'nodes': null,
    'edges': null
};
NodeProp.levelZeroNode = null;


var NodeGraph = {}; //store methods for the graph

/* Initialize node graph and append svg to body
 * Params: id of dom element to append svg
 */
NodeGraph.init = function(id, data, score) {
    //console.log(data);

    NodeGraph.setEvidenceScore(score); //set node score value
    NodeGraph.setGraphData(data);

    //save level zero node (risk node) for resetting the graph
    data.nodes.forEach(function(d) {
        if (d.level == 0) {
            NodeProp.levelZeroNode = d;
        }
    });


    $('#' + id).addClass('nodegraph'); //add class NodeGraph

    NodeProp.width = $('#' + id).width();


    //filter table to show evidence for level zero node
    alertOvrView.chrtTblGen(NodeProp.levelZeroNode);

    //createGraph
    NodeGraph.createGraph(data);
}


/* Initialize setting for the node graph simulation
 * Params: data for graph
 */
NodeGraph.generateSimulation = function(data) {
    /**Hack for Insider Trading**/
    let session = JSON.parse(sessionStorage.getItem("alertMetaData"));

    /**Hack for Insider Trading**/
    //check if the data has level three nodes
    let hasLevelThreeNodes = NodeGraph.checkForLevelThreeNodes(data);

    NodeProp.height = 500;

    NodeProp.svg = d3.select('#alrt-reasoning').append('svg').attr('width', NodeProp.width).attr('height', NodeProp.height)
        .attr("viewBox", "0 0 " + NodeProp.width + " " + NodeProp.height)
        .attr("preserveAspectRatio", "xMidYMid meet");


    //deselect any nodes when the graph is not clicked
    NodeProp.svg.on('click', function() {
        //console.log('reset graph');
        NodeGraph.deSelectNodes();
    });

    //add arrow to dom
    NodeProp.svg
        .append("defs")
        .append("marker")
        .attr("id", "arrow-lvl1")
        .attr('stroke-width', '4px')
        .attr("viewBox", "0 -5 10 10")
        .attr("refX", 38)
        .attr("refY", 0)
        .attr("markerWidth", 12)
        .attr("markerHeight", 12)
        .attr("orient", "auto")
        .attr("fill", "#D6D6D6")
        .append("svg:path")
        .attr("d", "M0,-5L10,0L0,5");

    NodeProp.svg
        .append("defs")
        .append("marker")
        .attr("id", "arrow-lvl2")
        .attr('stroke-width', '4px')
        .attr("viewBox", "0 -5 10 10")
        .attr("refX", 47)
        .attr("refY", 0)
        .attr("markerWidth", 8)
        .attr("markerHeight", 12)
        .attr("orient", "auto")
        .attr("fill", "#D6D6D6")
        .append("svg:path")
        .attr("d", "M0,-5L10,0L0,5");

    let iteration = 1; //num of times to run simulation
    let graphStrength = .03 * data.edges.length; // determines how compact the graph should be
    //console.log(graphStrength);
    if (graphStrength >= .4) {
        graphStrength = .9;
    }

    //issue with wash trade graph
    if (data.edges.length <= 10) {
        iteration = 4;
        graphStrength = .18;
        NodeProp.height = 400;
    }

    //off market graph
    if (data.edges.length <= 5) {
        iteration = 4;
        graphStrength = .2;
    }

    //Insider Trading
    if (session && session.alert && session.alert.type.toLowerCase() == "insider trading") {
        iteration = 5;
        graphStrength = .3;
        agrrWidth = 3;
    }

    // //console.log(graphStrength, data.edges.length);

    //instantiate d3 simulation object
    NodeProp.simulation = d3
        .forceSimulation()
        .force(
            "link",
            d3
            .forceLink(data)
            .id(function(d) {
                return d.id;
            })
            .strength(graphStrength)
            .iterations(iteration)
            .distance(
                function(d) {

                    if (NodeProp.isDrillDown) {
                        return 130;
                    }

                    if (d.source.level == 1) {
                        return 170;
                    } else {
                        return 10;
                    }
                }
            )
        )
        .force("charge", d3.forceManyBody().strength(-700))
        .force("center", d3.forceCenter(NodeProp.width / 2.1, NodeProp.height / 1.8));

    if (!NodeProp.isDrillDown) {
        if (session.alert.alertid == "20160902") {
            NodeProp.simulation.force('x', d3.forceX().strength(0))
                .force("center", d3.forceCenter(NodeProp.width / 2, NodeProp.height / 8))
                .force('y', d3.forceY(function(d) {
                    return (d.level === 0) ? (NodeProp.height * 2) : 0
                }).strength(.1));
        } else {
            NodeProp.simulation.force('x', d3.forceX().strength(0))
                .force("center", d3.forceCenter(NodeProp.width / 2, NodeProp.height / 3))
                .force('y', d3.forceY(function(d) {
                    return (d.level === 0) ? (NodeProp.height * 2) : 0
                }).strength(.1));
        }


        //increase charge value if model has no level three nodes
        if (!hasLevelThreeNodes) {
            NodeProp.simulation.force("charge", d3.forceManyBody().strength(-3000));
        }
    }
}


/* Create graph
 * Params: data for graph
 */
NodeGraph.createGraph = function(gData) {
    let data = $.extend(true, {}, gData); //create copy of data
    //console.log(data);

    NodeGraph.generateSimulation(data); //generate simulation settings for force graph
    let link = NodeProp.svg
        .append("g")
        .attr("class", "links")
        .selectAll("line")
        .data(data.edges)
        .enter()
        .append("line");

    let node = NodeProp.svg
        .append("g")
        .attr("class", "nodes")
        .selectAll("g")
        .data(data.nodes)
        .enter()
        .append("g")
        .attr("class", function(d) {
            return 'node level-' + d.level;
        });

    let label = NodeProp.svg.append("g").attr("class", "labels").selectAll("g")
        .data(data.nodes)
        .enter().append("g");

    label.append("text")
        .attr("x", 14)
        .attr("y", ".31em")
        .attr("dy", "0")
        .style("font-size", "13px")
        .style("fill", "#4b4b4b")
        .style("font-family", "IBM Helvetica Roman")
        .text(function(d) {
            if (d.level == 1 || NodeProp.isDrillDown) {
                return d.name;
            }
        });

    label.selectAll("text").call(wrapText, 100);

    let circleBg = node.append("circle").attr("fill", "#ffffff").attr("stroke-width", '1.5px')
        .attr("stroke", '#fff').attr('class', 'bg-circle');

    let circle = node.append("circle").attr("stroke-width", '1.5px')
        .attr("stroke", '#fff').attr('class', 'main-circle');

    let text = node
        .append("text")
        .attr("x", 21)
        .attr("y", ".33em")
        .style("font-family", "sans-serif")
        .style("font-size", "18px")
        .text(function(d) {
            if (d.level == 0) {
                return d.name;
            }

            if (d.level == 1) {
                d3.select(this).attr("x", 16);
                return d.name.substr(0, 1);
            }
        });


    let drillIcon = NodeProp.svg.selectAll(".level-1").append("g").attr("class", "drill-icon-group");

    drillIcon.append("circle").attr("r", 15).attr("fill", "#ffffff").attr('class', 'drilldown');
    drillIcon.append("text").attr("class", "drill-icon").html(function() {
        if (NodeProp.isDrillDown) {
            return "-";
        } else {
            return "+";
        }
    });

    //click event handler for node
    node.on("click", function(d) {
        //console.log("clicked", d.id);
        d3.event.stopPropagation();

        alertOvrView.clearInterval();
        alertOvrView.chrtTblGen(d);

        if (d.level !== 1) {
            nodeSelected(d3.select(this), d);
        } else {
            let childNode = alertOvrView.getAllChildNodes(d);
            if (NodeProp.isDrillDown) {
                NodeGraph.drillup();
            } else {
                NodeGraph.drilldown(d, childNode);
                NodeProp.drillDownNode = d;
            }
        }
    });

    //tools tips for node
    var tooltip = d3.select("body").append("div").attr("class", "nodegraph-toolTip");
    node.on("mouseenter", function(d) {

            var circle = d3.select(this).select('circle.main-circle');

            //toggle border to circle hover
            circleHoverState(d3.select(this), true);

            //get parameters to position tooltip
            var cx = circle.attr('cx');
            var cy = circle.attr('cy');
            var offSetY = parseInt(circle.attr('r')) * .9;
            var matrix = circle.node().getScreenCTM()
                .translate(+cx, +cy);

            //show tool tip
            if (d.level >= 2 && !NodeProp.isDrillDown) {
                tooltip.html('<span>' + d.name + '</span>');
                var offSetX = tooltip.style('width').replace('px', '') / 2;
                tooltip
                    .style("left", (window.pageXOffset + matrix.e - offSetX) + "px")
                    .style("top", (window.pageYOffset + matrix.f - (50 + offSetY)) + "px")
                    .style("opacity", "1");
            }
        })
        .on("mouseleave", function(d) {
            //toggle border to circle hover
            circleHoverState(d3.select(this), false);

            tooltip.style("opacity", "0");
        });


    NodeProp.simulation.nodes(data.nodes).on("tick", ticked);

    NodeProp.simulation.force("link").links(data.edges);

    NodeProp.simulation.on("end", function() {
        d3
            .select("svg")
            .transition()
            .duration(1000)
            .ease(d3.easeSin)
            .style("opacity", 1);
        // layout is done
        //console.log("on-end, graph done", link.data());
        if (!NodeProp.isInitialStateGenerated) {
            NodeProp.initialState.nodes = node.data();
            NodeProp.initialState.edges = link.data();
            NodeProp.isInitialStateGenerated = true;
        }

    });


    function ticked() {
        link
            .attr("x1", function(d) {
                return d.source.x;
            })
            .attr("y1", function(d) {
                return d.source.y;
            })
            .attr("x2", function(d) {
                return d.target.x;
            })
            .attr("y2", function(d) {
                return d.target.y;
            })
            .attr("marker-end", function(d) {
                if (d.source.level == 1) {
                    return "url(#arrow-lvl1)";
                }
                return "url(#arrow-lvl2)";
            });

        circle
            .attr("r", function(d) {
                if (d.level == 0) {
                    return 35;
                } else if (d.level == 1) {
                    return 30;
                } else {
                    return 7.5;
                }
            })
            .attr("cx", function(d) {
                d.fx = d.x;
                return d.x;
            })
            .attr("cy", function(d) {
                d.fy = d.y;
                return d.y;
            })
            .attr("fill", function(d) {
                var color = '';
                let col = NodeProp.evidenceScore.get(d.id) ? NodeProp.evidenceScore.get(d.id) : {
                    "score": 0
                };

                if (col.score > 0.75) {
                    color = "rgb(220, 0, 0)";
                } else if (col.score > 0.33 && col.score <= 0.75) {
                    color = "rgb(253, 140, 0)";
                } else if (col.score == 0) {
                    color = "rgb(119, 118, 119)";
                } else {
                    color = "rgb(253, 197, 0)";
                }

                return color;
            });


        circleBg.attr("r", function(d) {
                if (d.level == 0) {
                    return 35;
                } else if (d.level == 1) {
                    return 30;
                } else {
                    return 7.5;
                }
            }).attr("cx", function(d) {
                return d.x;
            })
            .attr("cy", function(d) {
                return d.y;
            });

        text.attr("transform", function(d) {
            if (d.level == 0) {
                return "translate(" + (d.x - 40) + "," + d.y + ")";
            }
            return "translate(" + (d.x - 22) + "," + d.y + ")";
        }).attr("fill", function(d) {
            let col = NodeProp.evidenceScore.get(d.id);
            if (col && col.score <= 0.33 && col.score != 0) {
                return 'black';
            }
            return 'white';
        });

        label.attr("transform", function(d) {
            if (d.level == 1) {
                return "translate(" + (d.x - 33) + "," + (d.y + 50) + ")";
            } else {
                return "translate(" + (d.x + 12) + "," + d.y + ")";
            }
        });

        label.selectAll("tspan").each(function(d, i) {
            d3.select(this).attr("y", (.28 * i) + 'em');
        });

        drillIcon.select("circle")
            .attr("cx", function(d) {
                return d.x;
            })
            .attr("cy", function(d) {
                return d.y;
            });

        drillIcon.select("text")
            .attr("x", function(d) {
                if (NodeProp.isDrillDown) {
                    return d.x - 4;
                } else {
                    return d.x - 6;
                }
            })
            .attr("y", function(d) {
                if (NodeProp.isDrillDown) {
                    return d.y + 6;
                } else {
                    return d.y + 5;
                }
            })
    }

    //stop graph from animating simulations
    //static source - https://bl.ocks.org/mbostock/1667139
    for (var i = 0,
            n = Math.ceil(Math.log(NodeProp.simulation.alphaMin()) / Math.log(1 - NodeProp.simulation.alphaDecay())); i < n;
        ++i
    ) {
        NodeProp.simulation.tick();
    }

    //toggle hover and active state for circle
    function circleHoverState(selectedNode, hover) {
        if (hover) {
            selectedNode.select('circle.main-circle').transition().style('opacity', '1').style('stroke', '#272727');
            selectedNode.select('text').transition().style('opacity', '1');
            selectedNode.select('.drill-icon-group').select('text').transition().style('opacity', '1');

        } else {
            if (selectedNode.attr('data-node-state') == 'inactive') {
                selectedNode.select('circle.main-circle').transition().style('stroke', 'white').style('opacity', '.3');
                selectedNode.select('.drill-icon-group').select('text').transition().style('opacity', '.3');
                if (selectedNode.select('circle.main-circle').attr('fill') == "rgb(253, 197, 0)") {
                    selectedNode.select('text').transition().style('opacity', '.3');
                }
            } else {
                selectedNode.select('circle.main-circle').transition().style('stroke', 'white');
            }

            if (selectedNode.attr('data-node-selected') == 'true') {
                selectedNode.select('circle.main-circle').transition().style('stroke', '#272727');
            }
        }
    }

    function nodeSelected(selectedNode, data) {
        let childNodes = alertOvrView.getAllChildNodes(data);

        //set propert for inactive nodes
        NodeProp.svg.selectAll('g.nodes g.node').each(function(node) {
            let currentNode = d3.select(this);
            let isChildNode = false;

            childNodes.forEach(function(childId) {
                if (childId == node.id) {
                    isChildNode = true;
                }
            });

            if (!isChildNode) {
                currentNode.attr('data-node-state', 'inactive').attr('data-node-selected', 'false');
                currentNode.select('circle.main-circle')
                    .transition().style('opacity', '.3')
                    .style('stroke', 'white');
                currentNode.select('.tsoTxt').transition().style('opacity', '.3');
                currentNode.select('.drill-icon-group').select('text').transition().style('opacity', '.3');

                //lower opacity for dark colors in the circle
                if (currentNode.select('circle.main-circle').attr('fill') == "rgb(253, 197, 0)") {
                    currentNode.select('text').transition().style('opacity', '.3');
                }
            } else {
                currentNode.attr('data-node-state', 'active').attr('data-node-selected', 'false');
                currentNode.select('circle.main-circle')
                    .transition().style('opacity', '1')
                    .style('stroke', 'white');
                currentNode.select('text').transition().style('opacity', '1');
                currentNode.select('.drill-icon-group').select('text').transition().style('opacity', '1');
            }
        });

        //set property for clicked node
        selectedNode.attr('data-node-state', 'active');
        selectedNode.attr('data-node-selected', 'true');
        selectedNode.select('circle.main-circle').transition().style('opacity', '1').style('stroke', '#272727');
        selectedNode.select('text').transition().style('opacity', '1');
        selectedNode.select('.drill-icon-group').select('text').transition().style('opacity', '1');
    }

    //wrap label text
    function wrapText(text, width) {
        text.each(function() {
            var text = d3.select(this),
                words = text.text().split(/\s+/).reverse(),
                word,
                line = [],
                lineNumber = 0,
                lineHeight = 1.1, // ems
                y = text.attr("y"),
                dy = parseFloat(text.attr("dy")),
                tspan = text.text(null).append("tspan").attr("x", 0).attr("y", y).attr("dy", dy + "em");
            while (word = words.pop()) {
                line.push(word);
                tspan.text(line.join(" "));
                if (tspan.node().getComputedTextLength() > width) {
                    line.pop();
                    tspan.text(line.join(" "));
                    line = [word];
                    tspan = text.append("tspan").attr("x", 0).attr("y", y).attr("dy", ++lineNumber * lineHeight + dy + "em").text(word);
                }
            }
        });
    }
}

/* Update score value for graph
 * Params: score: array
 */
NodeGraph.updateGraph = function(score) {
    //console.log('SCORE - ', score);
    NodeGraph.setEvidenceScore(score);
    NodeProp.svg.selectAll('g.nodes g').each(function(d) {
        var currNode = d3.select(this);

        //update circle color
        currNode.select('circle.main-circle')
            .transition()
            .ease(d3.easeCubic)
            .duration(1000)
            .attrTween("fill", function(d) {
                var color = getComputedStyle(this).getPropertyValue("fill");
                var oldColor = color;

                let col = NodeProp.evidenceScore.get(d.id) ? NodeProp.evidenceScore.get(d.id) : {
                    "score": 0
                };


                if (col.score > 0.75) {
                    color = "rgb(220, 0, 0)";
                } else if (col.score > 0.33 && col.score <= 0.75) {
                    color = "rgb(253, 140, 0)";
                } else if (col.score == 0) {
                    color = "rgb(119, 118, 119)";
                } else {
                    color = "rgb(253, 197, 0)";
                }

                if (oldColor != color) {
                    animateCircle(color);
                }

                return d3.interpolateRgb(oldColor, color);
            });


        //update text color
        currNode.select('text').transition().attr("fill", function(d) {
            let col = NodeProp.evidenceScore.get(d.id);
            if (col && col.score <= 0.33 && col.score != 0) {
                return 'black';
            }
            return 'white';
        });



        function animateCircle(fill) {
            var radius = parseInt(currNode.select('circle.main-circle').attr('r'));
            var cy = currNode.select('circle.main-circle').attr('cy');
            var cx = currNode.select('circle.main-circle').attr('cx');

            var circle = currNode
                .insert('circle', ":first-child")
                .attr('r', radius)
                .attr('fill', 'white')
                .attr('stroke', fill)
                .attr("stroke-width", '1.5px')
                .attr('cy', cy)
                .attr('cx', cx)
                .attr('opacity', '1');

            circle.transition()
                .ease(d3.easeSin)
                .delay(100)
                .duration(500)
                .attr("stroke-width", '30px')
                .attr('opacity', '.15')
                .transition()
                .attr('opacity', '.05')
                .duration(500)
                .attr("r", (radius + 20))
                .attr("stroke-width", '1px')
                .on("end", function(d) {
                    circle.remove();
                });
        }
    });
}

/* Set evidence score value
 * Params: score: array
 */
NodeGraph.setEvidenceScore = function(score) {
    NodeProp.evidenceScore= d3.map(score.values, function(d) {
        return d.nodeid;
    });
}

/* Set graph data
 * Params: data
 */
NodeGraph.setGraphData = function(data) {
    NodeProp.graphData = data;
}

/* Filter graph data for drill down and up feature
 * Params:
 * parent =  parent node
 * child: array = child nodes of parent node
 */
NodeGraph.filterGraphData = function(parent, child) {
    //console.log(NodeProp.graphData, parent, child);

    //filter edges based on parent id
    let edges = NodeProp.graphData.edges.filter(function(d) {
        if (d.target == parent.id) {
            return true;
        } else {
            return false;
        }
    });

    //filter nodes based on child nodes and parent nodes
    child.push(parent.id);
    let nodes = NodeProp.graphData.nodes.filter(function(d) {
        let found = false;
        child.forEach(function(e) {
            if (d.id == e) {
                found = true;
            }
        });

        return found;
    });

    let data = {
        'edges': edges,
        'nodes': nodes
    }

    return data;
}

/* Deselect any selected nodes
 * Params: none
 */
NodeGraph.deSelectNodes = function() {
    //reset nodes
    var i = 0;
    NodeProp.svg.selectAll('g.nodes g.node').each(function(d) {
        var currNode = d3.select(this);

        currNode.attr('data-node-state', null)
            .attr('data-node-selected', null);

        currNode.select('circle.main-circle')
            .transition()
            .duration(400)
            .delay(i * 20)
            .style('stroke', 'white')
            .style('opacity', '1');

        currNode.select('text').transition().style('opacity', '1');
        currNode.select('.drill-icon-group').select('text')
            .transition()
            .duration(200)
            .delay(i * 20)
            .style('opacity', '1');
        i++;
    });

    //update evidence table
    if (NodeProp.isDrillDown) {
        alertOvrView.chrtTblGen(NodeProp.drillDownNode);
    } else {
        alertOvrView.chrtTblGen(NodeProp.levelZeroNode);
    }
}

/* Drill down to level 1 nodes
 * Params:
 * parent =  parent node
 * child: array = child nodes of parent node
 */
NodeGraph.drilldown = function(parent, child) {
    let graphData = NodeGraph.filterGraphData(parent, child);
    d3.select('#alrt-reasoning').select('*').remove();
    NodeProp.isDrillDown = true;
    NodeGraph.createGraph(graphData);
}

/* Drill up // reset the graph
 */
NodeGraph.drillup = function() {
    d3.select('#alrt-reasoning').select('*').remove();
    NodeProp.isDrillDown = false;

    NodeGraph.createGraph(NodeProp.initialState);

    //reset table
    alertOvrView.chrtTblGen(NodeProp.levelZeroNode);
}

/* Check if the graph has level three nodes
 * The charge value for the simulation has to be increased for spoofing model
 */
NodeGraph.checkForLevelThreeNodes = function(data) {
    let found = false //boolean

    data.nodes.forEach(function(d) {
        if (d.level == 2) {
            found = true;
        }
    });

    return found;
}

/* Check if the graph has level three nodes
 * Set the graph level to zero
 */
NodeGraph.zeroLevel = function() {
    NodeGraph.setEvidenceScore({value: null});
    NodeProp.svg.selectAll('g.nodes g').each(function(d) {
        var currNode = d3.select(this);
        //update circle color
        currNode.select('circle.main-circle')
            .transition()
            .ease(d3.easeCubic)
            .duration(500)
            .attr("fill", "rgb(119, 118, 119)");
        });
}
